/* 31 March 2020, Ruud Harmsen.

   Display a file, conditional on the country code that belongs
   to the IP address of the visitor. If it is US, the file is
   not displayed, but instead an information screen, explaining
   about copyrights on works created before 1978, but published
   only after that. They seem to be extended in the USA, until
   31 December 2047. Public domain status for some other countries
   is doubtful too.

   17 July 2022:
   Made the program future-ready, so that after 31-12-2047 the
   copyright restriction is lifted. Hopefully by then the website
   is still up, including this program.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "../cgitools/cgitools.h"
#include "../cgitools/IP2cntr.h"

struct progdata_s
{
   FILE  *errfp;
   char  htmlfile[8+1+3+1];
   char  *scriptname;
   short force;
} progdata;

void get_invocation_parms (struct progdata_s *p_progdata);

void showform (struct progdata_s *p_progdata);

static char ErrorMessageBuffer[512];
static void ErrorOnScreen (void)
{
   printf("Content-type: text/html\n\n");
   printf("<html>\n");
   printf("<p>%s</p>\n", ErrorMessageBuffer);
   printf("</html>\n");

   fflush(stdout);
}

static char *ErrFile = "19782047/19782047.err";
int main (int argc, char **argv)
{
   int len;
   char *ErrFileMode = "at";
   progdata.scriptname = "../cgi-bin/19782047.cgi";

   progdata.errfp = fopen(ErrFile, ErrFileMode);
   if (!progdata.errfp)
   {
       strncpy(ErrorMessageBuffer, strerror(errno), sizeof ErrorMessageBuffer - 1);
       sprintf(ErrorMessageBuffer + strlen(ErrorMessageBuffer),
          "<br>Cannot open file \"%s\" in mode \"%s\", <br>user/grp = %d/%d, eff. user/grp = %d/%d\n",
          ErrFile, ErrFileMode, getuid(), geteuid(), getgid(), getegid());
       ErrorOnScreen();
       return 1;
   }

   get_invocation_parms(&progdata);
   if ((len = userinput()) == 0)
   {
      showform(&progdata);
   }

   fclose(progdata.errfp);

   return 0;
}

char *ForbiddenFilenames[] = {"copyriht.htm", "19782047.err", ".", "..", NULL};

void get_invocation_parms (struct progdata_s *p_progdata)
{
   char *query;
   char **Vars = NULL, **V;
   char **Ff;

   strcpy(p_progdata->htmlfile, "-");
   p_progdata->force = 0;

   query = getenv("QUERY_STRING");
   if (query)
   {
      Vars = CgiConvVars(query, NULL);

      for (V = Vars; V[0]; V += 2)
      {
         if (strcmp(V[0], "htmlfile") == 0 && strlen(V[1]))
         {
            sprintf(p_progdata->htmlfile, "%.*s",
               (int)(sizeof p_progdata->htmlfile) - 1, V[1]);
         }
         else if (strcmp(V[0], "force") == 0 && strlen(V[1]))
         {
            p_progdata->force = !strcmp(V[1], "0") ? 0 : 1;
         }
      }
   }

   for (Ff = ForbiddenFilenames; *Ff; Ff++)
   {
      if (strcmp(p_progdata->htmlfile, *Ff) == 0)
      {
         time_t now;
         fprintf(p_progdata->errfp, "%s", ctime(((now = time(NULL)), &now)));
         fprintf(p_progdata->errfp,
            "line = %d, Forbidden: p_progdata->htmlfile = \"%s\"\n",
            __LINE__, p_progdata->htmlfile);
         strcpy(p_progdata->htmlfile, "Forbidden");
      }
   }
   if (strchr(p_progdata->htmlfile, '/'))
   {
      time_t now;
      fprintf(p_progdata->errfp, "%s", ctime(((now = time(NULL)), &now)));
      fprintf(p_progdata->errfp,
         "line = %d, Forbidden: p_progdata->htmlfile = \"%s\"\n",
         __LINE__, p_progdata->htmlfile);
      strcpy(p_progdata->htmlfile, "Forbidden");
   }
}

void showform (struct progdata_s *p_progdata)
{
   char ISO_country[3];
   char htmlfile[41];
   FILE *fp;
   char c;
   int status;
   int after_2047;

   {
      /* Make the program future-ready, so that after 31-12-2047 the
         copyright restriction is lifted. Hopefully by then the website
         is still up, including this program. */
      struct tm then;
      time_t now = time(NULL);
      time_t t2048;

      then.tm_year  = 2047 - 1900;
      then.tm_mon   = 12 - 1;
      then.tm_mday  = 31;
      then.tm_hour  = 23;
      then.tm_min   = 59;
      then.tm_sec   = 59;
      then.tm_isdst = -1;

      after_2047 = now > (t2048 = mktime(&then));

      /*
      printf("Content-type: text/html\n\n");
      printf("<p>t2048 - tnow = %lld - %lld = %lld</p>\n",
         (long long)t2048, (long long)now, (long long)t2048 - now);
      printf("<p>After 2047 = %s</p>\n", after_2047 ? "yes" : "no");
      */
   }

   /* Exceptions for local test purposes: always allowed, so don't
      find the country code, assume the Netherlands. */
   strcpy(ISO_country, "NL");
   status = 0;
   if (strncmp(getenv("REMOTE_ADDR"), "127."    , 4) != 0 &&
       strncmp(getenv("REMOTE_ADDR"), "192.168.", 8) != 0)
   {
      /* Not local. So find country code. */
      status = IP2cntr(getenv("REMOTE_ADDR"), ISO_country, p_progdata->errfp);
      if (status != 0)
      {
         time_t now;
         fprintf(p_progdata->errfp, "%s line %d %s",
            __FILE__, __LINE__, ctime(((now = time(NULL)), &now)));
         fprintf(p_progdata->errfp, "Cannot determine country from IP address %s\n",
            getenv("REMOTE_ADDR"));

         sprintf(ErrorMessageBuffer, "See error in error file %s", ErrFile);
         ErrorOnScreen();

         return;
      }
   }
   if (p_progdata->force ||
      (!after_2047 && status == 0 && (
      strcmp(ISO_country, "US") == 0 ||
      strcmp(ISO_country, "CA") == 0 ||
      strcmp(ISO_country, "GB") == 0 ||
      strcmp(ISO_country, "IM") == 0 || /* Isle of Man */
      0)))
   {
      strcpy(htmlfile, "19782047/copyriht.htm");
   }
   else
   {
      sprintf(htmlfile, "19782047/%s", p_progdata->htmlfile);
   }

   fp = fopen(htmlfile, "rt");
   if (!fp)
   {
      time_t now;
      fprintf(p_progdata->errfp, "%s line %d %s",
          __FILE__, __LINE__, ctime(((now = time(NULL)), &now)));
      fprintf(p_progdata->errfp, "Cannot open file \"%s\", errno=%d = \"%s\"\n",
         htmlfile, errno, strerror(errno));

      sprintf(ErrorMessageBuffer, "See error in error file %s", ErrFile);
      ErrorOnScreen();
   }
   else
   {
      printf("Content-type: text/html\n\n");
      while ((c = getc(fp)) != EOF)
      {
         putchar(c);
      }
      fclose(fp);
   }
   fflush(stdout);
}
